/*
 * $Id: //depot/prod/orion/sc_main/mintaka/spl/snapcenter-cli/src/main/java/com/netapp/snapcenter/cli/command/AbstractSMSCommand.java#1 $
 *
 * Copyright (c) 2016 NetApp, Inc.
 * All rights reserved.
 */

package com.netapp.snapcenter.cli.command;

import com.beust.jcommander.Parameter;
import com.netapp.common.logging.MessageBundle;
import com.netapp.smcore.loader.logger.ExtendedGenericLogger;
import com.netapp.smcore.loader.logger.LoggerFactory;
import com.netapp.smcore.plugin.contracts.SMResult;
import com.netapp.smcore.plugin.contracts.SmRequest;
import com.netapp.smcore.plugin.contracts.SmResponse;
import com.netapp.snapcenter.cli.command.mapper.CommandMapperFactory;
import com.netapp.snapcenter.cli.command.mapper.CommandMapperInterface;
import com.netapp.snapcenter.cli.command.output.formatter.CommandOutputFormatter;
import com.netapp.snapcenter.cli.exception.CLIException;
import com.netapp.snapcenter.cli.formatter.Formatter;
import com.netapp.snapcenter.cli.formatter.FormatterFactory;
import com.netapp.snapcenter.cli.formatter.FormatterType;
import com.netapp.snapcenter.cli.helper.SnapCenterServerServiceFactory;
import com.netapp.snapcenter.cli.helper.SnapCenterServerServiceHelper;
import com.netapp.snapcenter.cli.logging.CLIMsgKey;
import com.netapp.snapcenter.cli.utils.CLIUtils;
import com.netapp.snapcenter.cli.utils.CLIUtils.LogLevel;
import com.netapp.snapcenter.cli.utils.ProgressBar;
import com.netapp.snapcenter.cli.utils.ProgressBarUtil;
import com.netapp.snapcenter.cli.utils.TokenStoreUtil;
import com.netapp.spl.common.restclient.context.SMSRestInvocationContext;
import org.apache.commons.lang3.StringUtils;

/**
 *
 * This is the base class for All SnapCenterServer Command. The run method is
 * common for all SMS commands and contains set of modules in it. The modules
 * inside run method can be overridden in case of specific requirement for that
 * command.
 *
 * This class also contains some util methods required by other SMS commands.
 *
 */

/**
 * TODO: All our help should be generated from the file
 * SnapManagerCore/PSHelp/en-US/SnapManager.PSModule.dll-help.xml.
 */
public abstract class AbstractSMSCommand implements CommandInterface {

    protected static final int STORAGE_CONNECTION_TIMEOUT_IN_SECONDS = 60;

    private ExtendedGenericLogger logger = LoggerFactory
            .getExtendedClassAwareLogger(AbstractSMSCommand.class);

    @Parameter(names = { "-Verbose" }, descriptionKey = "cli.abstract.command.verbose.option.description", hidden = true)
    protected boolean verbose;

    @Parameter(names = { "-Format" }, descriptionKey = "cli.abstract.command.format.option.description", hidden = true)
    protected FormatterType formatterType = FormatterType.TEXT_TABLE;

    @Parameter(help = true, names = { "-Help", "-?" }, descriptionKey = "cli.abstract.command.help.option.description")
    protected boolean help;
    
    @Parameter(names = { "-SetConsoleOutputWidth" }, descriptionKey = "cli.abstract.commmand.setConsoleOutputWidth.option.description")
    protected Integer setConsoleOutputWidth;
    
	public Integer getSetConsoleOutputWidth() {
		return setConsoleOutputWidth;
	}

	public void setSetConsoleOutputWidth(Integer setConsoleOutputWidth) {
		this.setConsoleOutputWidth = setConsoleOutputWidth;
	}

	private CommandMapperInterface commandMapper;

    private SnapCenterServerServiceHelper snapCenterServerServiceHelper;

    public abstract String getCommandName();

    public void run() throws CLIException {

        initialize();

        /*
         * Process Command arguments to perform validation, setting defaults
         * etc..
         */
        processCommandArguments();

        /*
         * Checking if Console OutputWidth is passed as Command arguments then setting specified ConsoleWidth 
         */
        if (setConsoleOutputWidth != null) {
            CLIUtils.setConsoleWidth(setConsoleOutputWidth);
        }
        
        // Command to request mapping
        SmRequest smRequest = (SmRequest) getCommandMapper().map(this);

        SmResponse smResponse = null;

        ProgressBar progressBar = null;

        try {
            // Start Progress Bar
            progressBar = ProgressBarUtil
                    .createAndStartProgressBar((MessageBundle
                            .localizeMessage(CLIMsgKey.CLI_PROGRESSBAR_CONTACTING_SNAPCENTER_SERVER)));
            // Invoke Server rest end point
            SMSRestInvocationContext<?, ?> commandRestInvocationContext = getCommandSpecificSMSRestInvocationContext(smRequest);
            smResponse = (SmResponse) getSnapCenterServerServiceHelper()
                    .makeRestCallToSnapCenterServer(
                            commandRestInvocationContext);
        } catch (CLIException e) {
            throw e;
        } finally {
            // Stop Progress Bar
            ProgressBarUtil.stopProgressBar(progressBar);
        }

        // Command should override to poll for job status
        pollForJobStatus(smResponse);

        // Tasks like printing response etc.. are done here.
        processCommandResponse(smResponse);

    }

    protected void initialize() throws CLIException {
        if (snapCenterServerServiceHelper == null) {
            snapCenterServerServiceHelper = SnapCenterServerServiceFactory
                    .getSnapCenterServerServiceHelperInstance();
        }
        String token = TokenStoreUtil.readToken();
        if (!(this instanceof OpenSmConnectionCommand)
                && StringUtils.isBlank(token)) {
            String openSmConnectionCommandName = MessageBundle
                    .localizeMessage(CLIMsgKey.CLI_OPENSMCONNECTION_COMMAND_NAME);
            throw new CLIException(MessageBundle.localizeMessage(
                    CLIMsgKey.CLI_SERVER_CONNECTION_ERROR,
                    openSmConnectionCommandName));
        }
        snapCenterServerServiceHelper.setToken(token);
    }

    public CommandMapperInterface getCommandMapper() {
        if (commandMapper == null) {
            commandMapper = CommandMapperFactory.getCommandMapper();
        }
        return commandMapper;
    }

    protected void processCommandResponse(SmResponse smResponse)
            throws CLIException {
        // Prints the Command response Object Based on the formatter option used
        printCommandResponseObject(smResponse);

        // Prints the command result
        displayCommandResult(smResponse);

    }

    protected void processCommandArguments() throws CLIException {
        /*
         * Do nothing here. If a specific command need to process its argument
         * will override this method
         */
        return;
    }

    protected <T> void pollForJobStatus(SmResponse smResponse)
            throws CLIException {
        /*
         * Do nothing here. If a specific command need to have poll feature it
         * will override this method
         */
        return;
    }

    /**
     *
     * To print Command Response Object based on formatter and verbose mode.
     * Commands can override this to provide their specific objects printing.
     *
     */
    protected <T> void printCommandResponseObject(T smResponse)
            throws CLIException {
        CommandOutputFormatter cof = getCommandSpecificOutputFormatter();
        if (cof != null) {
            cof.formatAndPrintCommandResponseObject(smResponse, getFormatterType());
        } else {
            Formatter formatter = FormatterFactory
                    .getFormatter(getFormatterType());
            formatter.writeFormattedObject(smResponse);
        }
    }

    /**
     * Commands should override this method if they have their own formatters
     *
     */
    protected CommandOutputFormatter getCommandSpecificOutputFormatter() {
        return null;
    }
    /**
     * Specific command can overide this in case if there is any specific
     * message to be displayed.
     *
     * @param smResponse
     */
    protected void displayCommandResult(SmResponse smResponse)
            throws CLIException {

        SMResult smResult = smResponse.getResult();
        if (isCommandExecutionSuccessfull(smResult)) {
            String executionSucess = MessageBundle.localizeMessage(
                    CLIMsgKey.CLI_COMMAND_EXECUTION_SUCCESS,
                    this.getCommandName());
            CLIUtils.writeToLogAndConsole(executionSucess, logger,
                    LogLevel.INFO);
        } else {

            String executionFailed = String.format("%s %s", smResult
                    .getMessage(), MessageBundle.localizeMessage(
                    CLIMsgKey.CLI_SERVER_ERROR_CODE, smResult.getErrorCode()));
            throw new CLIException(executionFailed);
        }

    }

    protected boolean isCommandExecutionSuccessfull(SMResult smResult) {

        return getSnapCenterServerServiceHelper()
                .isSnapCenterRestOperationResponseResultSuccess(smResult);
    }

    protected abstract <T extends SmRequest> SMSRestInvocationContext<?, ?> getCommandSpecificSMSRestInvocationContext(
            T cmdRequest);

    protected void doPollForJobStatus(SmResponse smResponse,
                                      boolean waitForJobCompletion) throws CLIException {
        if (smResponse.getJob() == null) {
            throw new CLIException(
                    MessageBundle
                            .localizeMessage(
                                    CLIMsgKey.CLI_SERVER_UNEXPECTED_ERROR,
                                    MessageBundle
                                            .localizeMessage(CLIMsgKey.CLI_SERVER_JOB_DETAILS_RECEIVE_ERROR)));
        }

        if (!isCommandExecutionSuccessfull(smResponse.getResult())) {
            return;
        }

        String displayJobNameStatusId = MessageBundle.localizeMessage(
                CLIMsgKey.CLI_SERVER_JOB_NAME_STATUS_ID, smResponse.getJob()
                        .getName(), smResponse.getJob().getStatus(), smResponse
                        .getJob().getId());
        CLIUtils.writeToLogAndConsole(displayJobNameStatusId, logger,
                LogLevel.INFO);

        if (!waitForJobCompletion) {
            return;
        } else {
            getSnapCenterServerServiceHelper().pollSMSJobStatus(
                    smResponse.getJob().getId());
        }

    }

    public boolean isHelpSpecified() {
        return help;
    }

    public boolean isVerbose() {
        return verbose;
    }

    public void setHelp(boolean help) {
        this.help = help;
    }

    public void setVerbose(boolean verbose) {
        this.verbose = verbose;
    }

    public void setFormatterType(FormatterType formatterType) {
        this.formatterType = formatterType;
    }

    public FormatterType getFormatterType() {
        return formatterType;
    }

    public SnapCenterServerServiceHelper getSnapCenterServerServiceHelper() {
        return snapCenterServerServiceHelper;
    }

    public void setSnapCenterServerServiceHelper(
            SnapCenterServerServiceHelper snapCenterServerServiceHelper) {
        this.snapCenterServerServiceHelper = snapCenterServerServiceHelper;
    }
}
