/*
 * $Id: //depot/prod/spl/main/snapcenter-cli/src/main/java/com/netapp/snapcenter/cli/command/output/formatter/GetSmResourcesCommandOutputFormatter.java#1 $
 *
 * Copyright (c) 2016 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.snapcenter.cli.command.output.formatter;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.text.DecimalFormat;
import java.util.List;

import com.netapp.common.logging.MessageBundle;
import com.netapp.smcore.loader.logger.ExtendedGenericLogger;
import com.netapp.smcore.loader.logger.LoggerFactory;
import com.netapp.smcore.plugin.contracts.ArrayOfSmAppFileStorageGroup;
import com.netapp.smcore.plugin.contracts.ArrayOfSmObjectOperationResult;
import com.netapp.smcore.plugin.contracts.ArrayOfSmResourceResult;
import com.netapp.smcore.plugin.contracts.SDNFSMountPoint;
import com.netapp.smcore.plugin.contracts.SDStorageFootprint;
import com.netapp.smcore.plugin.contracts.SmAppFileStorageGroup;
import com.netapp.smcore.plugin.contracts.SmDiscoverResourcesResponse;
import com.netapp.smcore.plugin.contracts.SmObject;
import com.netapp.smcore.plugin.contracts.SmObjectOperationResult;
import com.netapp.smcore.plugin.contracts.SmOracleDb;
import com.netapp.smcore.plugin.contracts.SmResourceResult;
import com.netapp.smcore.plugin.contracts.SmSnapDriveResource;
import com.netapp.snapcenter.cli.command.GetSmResourcesCommand;
import com.netapp.snapcenter.cli.command.output.formatter.CommandOutputFormatter;
import com.netapp.snapcenter.cli.contracts.PluginCode;
import com.netapp.snapcenter.cli.contracts.ResourceType;
import com.netapp.snapcenter.cli.exception.CLIException;
import com.netapp.snapcenter.cli.formatter.Formatter;
import com.netapp.snapcenter.cli.formatter.FormatterFactory;
import com.netapp.snapcenter.cli.formatter.FormatterType;
import com.netapp.snapcenter.cli.formatter.TextTable;
import com.netapp.snapcenter.cli.formatter.TextTableInterface;
import com.netapp.snapcenter.cli.utils.CLIUtils;
import com.netapp.snapcenter.cli.utils.OracleDatabaseDetailsUtils;
import com.netapp.snapcenter.cli.utils.CLIUtils.LogLevel;
import com.netapp.snapcenter.cli.logging.CLIMsgKey;

/**
 * This class implements Get-SmResources output formatter, command output will
 * be processed here instead of handling it in the request class itself.
 */
public class GetSmResourcesCommandOutputFormatter implements
        CommandOutputFormatter {

    private PluginCode pluginCode;
    private static ExtendedGenericLogger logger = LoggerFactory
            .getExtendedClassAwareLogger(GetSmResourcesCommandOutputFormatter.class);
    private static final double TB = 1099511627776d;
    private static final double GB = 1073741824d;
    private static final double MB = 1048576d;
    private static final double KB = 1024d;

    public GetSmResourcesCommandOutputFormatter(PluginCode pluginCode) {
        this.pluginCode = pluginCode;
    }

    @Override
    public <T> void formatAndPrintCommandResponseObject(T smResponse,
            FormatterType formatterType) throws CLIException {
        SmDiscoverResourcesResponse smDiscoverResourcesResponse = (SmDiscoverResourcesResponse) smResponse;

        ArrayOfSmResourceResult arrayOfSmResourceResult = smDiscoverResourcesResponse
                .getResources();
        if (arrayOfSmResourceResult == null
                || CLIUtils.isNullOrEmptyCollection(arrayOfSmResourceResult
                        .getSmResourceResult())) {
            return;
        }

        switch (formatterType) {
            case TEXT:
                if (pluginCode.equals(PluginCode.SCU)) {
                    printFileSystemResourcesInTextFormat(arrayOfSmResourceResult);
                    break;
                }
            case JSON:
            case XML:
                Formatter formatter = FormatterFactory.getFormatter(formatterType);
                formatter.writeFormattedObject(smDiscoverResourcesResponse);
                break;
            case TEXT_TABLE:
            default:
                if (pluginCode.equals(PluginCode.SCO)) {
                    printOracleResources(arrayOfSmResourceResult);
                } else if (pluginCode.equals(PluginCode.SCU)) {
                    printFileSystemResources(arrayOfSmResourceResult);
                }
        }
    }

    /**
     * Prints the Oracle database resource on console
     * 
     * @param arrayOfSmResourceResult
     *            input, Array of SmResourceResult
     * 
     */
    private void printOracleResources(
            ArrayOfSmResourceResult arrayOfSmResourceResult) {
        String headerResourceName = MessageBundle
                .localizeMessage(CLIMsgKey.CLI_GETSMRESOURCES_TABLE_RESOURCE_NAME_HEADER);
        String headerResourceVersion = MessageBundle
                .localizeMessage(CLIMsgKey.CLI_GETSMRESOURCES_TABLE_RESOURCE_VERSION_HEADER);
        String headerResourceId = MessageBundle
                .localizeMessage(CLIMsgKey.CLI_GETSMRESOURCES_TABLE_RESOURCE_ID_HEADER);
        String headerResourceType = MessageBundle
                .localizeMessage(CLIMsgKey.CLI_GETSMRESOURCES_TABLE_RESOURCE_TYPE_HEADER);
        String headerResourceOverallStatus = MessageBundle
                .localizeMessage(CLIMsgKey.CLI_GETSMRESOURCES_TABLE_RESOURCE_OVERALL_STATUS_HEADER);
        String[] resourcesTableHeaders = { headerResourceName,
                headerResourceVersion, headerResourceId, headerResourceType, headerResourceOverallStatus};

        TextTableInterface resourcesTable = new TextTable(resourcesTableHeaders);
        for (SmResourceResult smResourceResult : arrayOfSmResourceResult
                .getSmResourceResult()) {

            ArrayOfSmObjectOperationResult arrayOfSmObjectOperationResult = smResourceResult
                    .getOperationResults();
            if (arrayOfSmObjectOperationResult == null
                    || CLIUtils
                            .isNullOrEmptyCollection(arrayOfSmObjectOperationResult
                                    .getSmObjectOperationResult())) {
                continue;
            }
            for (SmObjectOperationResult smObjectOperationResult : arrayOfSmObjectOperationResult
                    .getSmObjectOperationResult()) {
                SmOracleDb smOracleDb = (SmOracleDb) smObjectOperationResult
                        .getTarget();
                if (smOracleDb != null) {
                    String columnResourceName = CLIUtils
                            .getStringValue(smOracleDb.getName());
                    String columnResourceVersion = CLIUtils
                            .getStringValue(smOracleDb.getDbVersion());
                    String columnResourceId = CLIUtils
                            .getStringValue(smOracleDb.getId());
                    String columnResourceType = OracleDatabaseDetailsUtils
                            .getOracleDatabaseTypeForDisplay(smOracleDb);
                    String overallStatus = String.valueOf(smOracleDb.getLastBackupStatus());
                    resourcesTable.addRow(columnResourceName,
                            columnResourceVersion, columnResourceId,
                            columnResourceType,overallStatus);
                }

            }

        }
        resourcesTable.displayFormattedTable();

    }

    /**
     * Prints the file system resource on console
     * 
     * @param arrayOfSmResourceResult
     *            input, Array of SmResourceResult
     * 
     */
    private void printFileSystemResources(
            ArrayOfSmResourceResult arrayOfSmResourceResult) {

        String headerResourceName = MessageBundle
                .localizeMessage(CLIMsgKey.CLI_GETSMRESOURCES_TABLE_RESOURCE_NAME_HEADER);
        String headerResourceId = MessageBundle
                .localizeMessage(CLIMsgKey.CLI_GETSMRESOURCES_TABLE_RESOURCE_ID_HEADER);
        String headerResourceType = MessageBundle
                .localizeMessage(CLIMsgKey.CLI_GETSMRESOURCES_TABLE_RESOURCE_TYPE_HEADER);
        String headerResourceSize = MessageBundle
                .localizeMessage(CLIMsgKey.CLI_GETSMRESOURCES_TABLE_RESOURCE_SIZE_HEADER);
        String[] resourcesTableHeaders = { headerResourceName,
            headerResourceId, headerResourceType, headerResourceSize };
        TextTableInterface resourcesTable = new TextTable(resourcesTableHeaders);

        for (SmResourceResult smResourceResult : arrayOfSmResourceResult
                .getSmResourceResult()) {

            ArrayOfSmObjectOperationResult arrayOfSmObjectOperationResult = smResourceResult
                    .getOperationResults();
            if (arrayOfSmObjectOperationResult == null
                    || CLIUtils
                            .isNullOrEmptyCollection(arrayOfSmObjectOperationResult
                                    .getSmObjectOperationResult())) {
                continue;
            }
            for (SmObjectOperationResult smObjectOperationResult : arrayOfSmObjectOperationResult
                    .getSmObjectOperationResult()) {
                SmSnapDriveResource smSnapDriveResource = (SmSnapDriveResource) smObjectOperationResult
                        .getTarget();
                if (smSnapDriveResource != null) {

                    // TODO: In future need to handle it for SDSANMountPoint
                    String columnResourceName = CLIUtils
                            .getStringValue(smSnapDriveResource.getName());
                    String columnResourceType = CLIUtils
                            .getStringValue(smSnapDriveResource.getObjectType());
                    String host = CLIUtils.getStringValue(smSnapDriveResource
                            .getHost());
                    // TODO: After the server side changes it will read from the
                    // response object
                    String columnResourceId = host + "\\" + columnResourceName;
                    String columnResourceSize = calculateSize(smSnapDriveResource
                            .getObjectSize());
                    resourcesTable.addRow(columnResourceName, columnResourceId,
                            columnResourceType, columnResourceSize);
                }

            }

        }
        resourcesTable.displayFormattedTable();

    }

    private void printFileSystemResourcesInTextFormat(
            ArrayOfSmResourceResult arrayOfSmResourceResult) {

        StringBuilder commandOutput = new StringBuilder();

        for (SmResourceResult smResourceResult : arrayOfSmResourceResult
                .getSmResourceResult()) {

            ArrayOfSmObjectOperationResult arrayOfSmObjectOperationResult = smResourceResult
                    .getOperationResults();
            if (arrayOfSmObjectOperationResult == null
                    || CLIUtils
                            .isNullOrEmptyCollection(arrayOfSmObjectOperationResult
                                    .getSmObjectOperationResult())) {
                continue;
            }
            for (SmObjectOperationResult smObjectOperationResult : arrayOfSmObjectOperationResult
                    .getSmObjectOperationResult()) {
                SmSnapDriveResource smSnapDriveResource = (SmSnapDriveResource) smObjectOperationResult
                        .getTarget();
                if (smSnapDriveResource != null) {

                    // TODO: In future need to handle it for SDSANMountPoint
                    String resourceName = CLIUtils
                            .getStringValue(smSnapDriveResource.getName());
                    String resourceType = CLIUtils
                            .getStringValue(smSnapDriveResource.getObjectType());
                    String host = CLIUtils.getStringValue(smSnapDriveResource
                            .getHost());
                    // TODO: After the server side changes it will read from the
                    // response object
                    String resourceId = host + "\\" + resourceName;
                    String resourceSize = calculateSize(smSnapDriveResource
                            .getObjectSize());

                    commandOutput
                            .append(MessageBundle
                                    .localizeMessage(
                                            CLIMsgKey.CLI_GETSMRESOURCES_TEXTVIEW_RESOURCE_NAME,
                                            resourceName));
                    commandOutput.append(CLIUtils.NEW_LINE);
                    commandOutput.append(MessageBundle.localizeMessage(
                            CLIMsgKey.CLI_GETSMRESOURCES_TEXTVIEW_RESOURCE_ID,
                            resourceId));
                    commandOutput.append(CLIUtils.NEW_LINE);
                    commandOutput
                            .append(MessageBundle
                                    .localizeMessage(
                                            CLIMsgKey.CLI_GETSMRESOURCES_TEXTVIEW_RESOURCE_TYPE,
                                            resourceType));
                    commandOutput.append(CLIUtils.NEW_LINE);
                    commandOutput
                            .append(MessageBundle
                                    .localizeMessage(
                                            CLIMsgKey.CLI_GETSMRESOURCES_TEXTVIEW_RESOURCE_SIZE,
                                            resourceSize));
                    commandOutput.append(CLIUtils.NEW_LINE);
                    commandOutput.append(CLIUtils.NEW_LINE);
                }
            }
            CLIUtils.writeToConsole(commandOutput.toString());
        }
    }

    private String calculateSize(BigInteger filesystemSize) {
        double size = filesystemSize.doubleValue();
        String returnValue = new String();
        DecimalFormat format = new DecimalFormat("0.#");

        if (size >= TB) {
            size = (double) Math.round((size / TB) * 10) / 10;
            returnValue = format.format(size) + "T";
        } else if (size >= GB) {
            size = (double) Math.round((size / GB) * 10) / 10;
            returnValue = format.format(size) + "G";
        } else if (size >= MB) {
            size = (double) Math.round((size / MB) * 10) / 10;
            returnValue = format.format(size) + "M";
        } else if (size >= KB) {
            size = (double) Math.round((size / KB) * 10) / 10;
            returnValue = format.format(size) + "K";
        } else if (size < KB) {
            size = (double) Math.round((size * 10) / 10);
            returnValue = format.format(size) + "B";
        }
        return returnValue;
    }

}
