/*
 * $Id: //depot/prod/spl/main/snapcenter-cli/src/main/java/com/netapp/snapcenter/cli/helper/SnapCenterServerServiceHelper.java#1 $
 * 
 * Copyright (c) 2015 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.snapcenter.cli.helper;

import static com.netapp.spl.common.restclient.helper.PropertiesFileReader.SCCLI_CONFIG_SNAPCENTER_SERVER_JOB_STATUS_POLLING_TIME_INTERVAL_SECONDS;

import javax.ws.rs.core.UriBuilder;

import org.apache.commons.lang3.StringUtils;

import com.netapp.common.logging.MessageBundle;
import com.netapp.smcore.loader.logger.ExtendedGenericLogger;
import com.netapp.smcore.loader.logger.LoggerFactory;
import com.netapp.smcore.plugin.contracts.SMResult;
import com.netapp.smcore.plugin.contracts.SmAddStorageConnectionRequest;
import com.netapp.smcore.plugin.contracts.SmCreateRunAsAccountRequest;
import com.netapp.smcore.plugin.contracts.SmGetJobStatusRequest;
import com.netapp.smcore.plugin.contracts.SmGetJobStatusResponse;
import com.netapp.smcore.plugin.contracts.SmJob;
import com.netapp.smcore.plugin.contracts.SmJobStatus;
import com.netapp.smcore.plugin.contracts.SmLoginRequest;
import com.netapp.smcore.plugin.contracts.SmModifyRunAsAccountRequest;
import com.netapp.smcore.plugin.contracts.SmRequestBase;
import com.netapp.smcore.plugin.contracts.SmResponseBase;
import com.netapp.snapcenter.cli.exception.CLIException;
import com.netapp.snapcenter.cli.logging.CLIMsgKey;
import com.netapp.snapcenter.cli.utils.CLIUtils;
import com.netapp.snapcenter.cli.utils.CLIUtils.LogLevel;
import com.netapp.snapcenter.cli.utils.ProgressBar;
import com.netapp.snapcenter.cli.utils.ProgressBarUtil;
import com.netapp.snapcenter.cli.utils.SMSObjectSerializationUtil;
import com.netapp.spl.common.restclient.RestClientFactory;
import com.netapp.spl.common.restclient.RestClientInterface;
import com.netapp.spl.common.restclient.config.SMSRestConfiguration;
import com.netapp.spl.common.restclient.config.SMSRestConfiguration.SMSRestConfigurationBuilder;
import com.netapp.spl.common.restclient.context.GenericRestInvocationContext;
import com.netapp.spl.common.restclient.context.SMSRestInvocationContext;
import com.netapp.spl.common.restclient.exception.RestClientException;
import com.netapp.spl.common.restclient.helper.PropertiesFileReader;

/***
 * This class is for having SnapCenter Server related activity. It has an
 * utility for consuming snapcenter rest end point services, Polling server jobs
 * etc..
 *
 */
public class SnapCenterServerServiceHelper {

    private static final long DEFAULT_JOB_STATUS_POLLING_TIME_INTERVAL = 5 * 1000L;

    private String token;
    private String snapCenterHost;
    private int snapCenterPort;
    private String snapCenterProtocol;
    private long connectionTimeout;
    private long receiveTimeout;
    private String snapCenterBaseURL;

    private LogLevel requestAndResponseLogLevel = LogLevel.DEBUG;

    private ExtendedGenericLogger logger = LoggerFactory
            .getExtendedClassAwareLogger(SnapCenterServerServiceHelper.class);

    public SnapCenterServerServiceHelper(SnapCenterServerConfig snapCenterServerConfig) {
        this.snapCenterHost = snapCenterServerConfig.getHost();
        this.snapCenterPort = snapCenterServerConfig.getPort();
        this.snapCenterProtocol = snapCenterServerConfig.getProtocol();
        this.connectionTimeout = snapCenterServerConfig.getConnectionTimeout();
        this.receiveTimeout = snapCenterServerConfig.getRecieveTimeout();

        setSnapCenterBaseURL(UriBuilder.fromPath("/")
                .scheme(snapCenterProtocol).host(snapCenterHost)
                .port(snapCenterPort).build().toString());

    }

    public <T extends SmRequestBase, V extends SmResponseBase> V makeRestCallToSnapCenterServer(
            SMSRestInvocationContext<T, V> smsRestInvocationContext)
            throws CLIException {
        logSMSRequestObject(smsRestInvocationContext.getRequest());

        V smResponseBase = makeRestCallToSnapCenterServerWithGenericRestInvocationContext(smsRestInvocationContext);

        logSMSResponseObject(smResponseBase);

        return smResponseBase;
    }

    public <T extends Object, V extends Object> V makeRestCallToSnapCenterServerWithGenericRestInvocationContext(
                GenericRestInvocationContext<T, V> genericRestInvocationContext)
                throws CLIException {
        V response = null;
        try {

            SMSRestConfiguration smsRestConfiguration = new SMSRestConfigurationBuilder()
                    .setProtocol(snapCenterProtocol).setHost(snapCenterHost)
                    .setPort(snapCenterPort)
                    .setConnectionTimeout(connectionTimeout)
                    .setReceiveTimeout(receiveTimeout)
                    .build();

            /*
             * Another way of setting token. Should be used only when the
             * existing way has any problem or unavailable. Warning: Setting
             * token in this way will expose token if Request is logged.
             * 
             * 
             * SmUserOperationContext smUserOperationContext =
             * (smRequest.getUserOperationContext() == null ) ? new
             * SmUserOperationContext() : smRequest.getUserOperationContext();
             * SmUser smUser = (smUserOperationContext.getUser() == null) ? new
             * SmUser() : smUserOperationContext.getUser();
             * smUser.setToken(smsToken);
             * 
             * smUserOperationContext.setUser(smUser);
             * smRequest.setUserOperationContext(smUserOperationContext);
             */
            smsRestConfiguration.setToken(getToken());

            RestClientInterface restClient = RestClientFactory.getRestClient();
            restClient.setRestConfiguration(smsRestConfiguration);

            response =  restClient.invokeRestEndPoint(genericRestInvocationContext);

        } catch (RestClientException e) {
            CLIException cliException = new CLIException(e.getLocalizedMessage());
            cliException.setResponseStatusCode(e.getResponseStatusCode());
            cliException.setResponseStatusMessage(e.getResponseStatusMessage());
            throw cliException;
        }

        return response;
    }

    public void pollSMSJobStatus(Long smJobId) throws CLIException {
 
        long jobStatusPollingTimeInterval = DEFAULT_JOB_STATUS_POLLING_TIME_INTERVAL;

        try {
            String jobStatusPollValueInSec = PropertiesFileReader
                    .getInstance()
                    .getProperty(
                            SCCLI_CONFIG_SNAPCENTER_SERVER_JOB_STATUS_POLLING_TIME_INTERVAL_SECONDS);
            jobStatusPollingTimeInterval = Long
                    .parseLong(jobStatusPollValueInSec) * 1000L;
        } catch (Exception e) {
            CLIUtils.writeToLog(e, logger, LogLevel.DEBUG);
        }

        SmGetJobStatusRequest smGetJobStatusRequest = new SmGetJobStatusRequest();
        smGetJobStatusRequest.setJobId(smJobId);
        smGetJobStatusRequest.setError(true);

        SMSRestInvocationContext<SmGetJobStatusRequest, SmGetJobStatusResponse> smsRestInvocationContext = new SMSRestInvocationContext<>();
        smsRestInvocationContext.setRequest(smGetJobStatusRequest);
        smsRestInvocationContext
                .setResponseClass(SmGetJobStatusResponse.class);
        smsRestInvocationContext
                .setServiceUrlPath(SMSRestConfiguration.SMS_GET_JOBSTATUS_SERVICE_PATH);

        // For polling request keep log level to trace
        LogLevel logLevelBeforePolling = getRequestAndResponseLogLevel();
        setRequestAndResponseLogLevel(LogLevel.TRACE);

        ProgressBar progressBar = ProgressBarUtil
                .createAndStartProgressBar(String.format(
                        "Waiting for completion of job %s ",
                        Long.toString(smJobId.longValue())));

        SmJob smJob = null;
        do {
            try {
                Thread.sleep(jobStatusPollingTimeInterval);
            } catch (InterruptedException e) {
                throw new CLIException(e.getLocalizedMessage());
            }

            SmGetJobStatusResponse smGetJobStatusResponse = makeRestCallToSnapCenterServer(smsRestInvocationContext);
            smJob = smGetJobStatusResponse.getJobs().getSmJob().iterator().next();
        } while (smJob != null
                && (smJob.getStatus().equals(SmJobStatus.RUNNING) || smJob
                        .getStatus().equals(SmJobStatus.QUEUED)));
       
        ProgressBarUtil.stopProgressBar(progressBar);

        // Once polling is done set it back to previous value
        setRequestAndResponseLogLevel(logLevelBeforePolling);

        String jobStatus = MessageBundle.localizeMessage(
                CLIMsgKey.CLI_SERVER_JOB_STATUS, smJob.getStatus().value());

        if (!StringUtils.isEmpty(smJob.getError())) {
            String jobError = String.format("%s: %s", jobStatus,
                    smJob.getError());
            throw new CLIException(jobError);
        } else {
            CLIUtils.writeToLogAndConsole(jobStatus.toString(), logger,
                    LogLevel.INFO);
        }
    }

    public boolean isSnapCenterRestOperationResponseResultSuccess(
            SMResult smResult) {
        boolean success = true;
        if (smResult.getErrorCode() != 0
                || StringUtils.isNotEmpty(smResult.getMessage())) {
            success = false;
        }
        return success;
    }

    public <T extends SmRequestBase, V extends SmResponseBase> SMSRestInvocationContext<?, ?> createSMSRestInvocationContext(
            T t, Class<V> v, String servicePath) {
        SMSRestInvocationContext<T, V> smsRestInvocationContext = new SMSRestInvocationContext<>();
        smsRestInvocationContext.setRequest(t);
        smsRestInvocationContext.setResponseClass(v);
        smsRestInvocationContext.setServiceUrlPath(servicePath);
        return smsRestInvocationContext;
    }

    private void logSMSRequestObject(SmRequestBase smRequestBase) {
        if (smRequestBase instanceof SmLoginRequest
                || smRequestBase instanceof SmAddStorageConnectionRequest
                || smRequestBase instanceof SmCreateRunAsAccountRequest
                || smRequestBase instanceof SmModifyRunAsAccountRequest) {
            /*
             * Don't log these requests since it contains the raw passwords.
             * TODO: Remove this constraint when password is encrypted. For now
             * it's NO Op
             */
            return;
        }
        CLIUtils.writeToLog(
                "SMS Request Object: \n"
                        + SMSObjectSerializationUtil
                                .getXMLSerializedSMSObject(smRequestBase),
                logger, getRequestAndResponseLogLevel());
    }

    private void logSMSResponseObject(SmResponseBase smResponseBase) {
        CLIUtils.writeToLog(
                "SMS Response Object: \n"
                        + SMSObjectSerializationUtil
                                .getXMLSerializedSMSObject(smResponseBase),
                logger, getRequestAndResponseLogLevel());
    }

    public void setToken(String token) {
        this.token = token;
    }

    public String getToken() {
        return token;
    }

    public void setRequestAndResponseLogLevel(
            LogLevel requestAndResponseLogLevel) {
        this.requestAndResponseLogLevel = requestAndResponseLogLevel;
    }

    public LogLevel getRequestAndResponseLogLevel() {
        return requestAndResponseLogLevel;
    }

    public String getSnapCenterHost() {
        return snapCenterHost;
    }

    public void setSnapCenterHost(String snapCenterHost) {
        this.snapCenterHost = snapCenterHost;
    }

    public int getSnapCenterPort() {
        return snapCenterPort;
    }

    public void setSnapCenterPort(int snapCenterPort) {
        this.snapCenterPort = snapCenterPort;
    }

    public String getSnapCenterProtocol() {
        return snapCenterProtocol;
    }

    public void setSnapCenterProtocol(String snapCenterProtocol) {
        this.snapCenterProtocol = snapCenterProtocol;
    }

    public long getConnectionTimeout() {
        return connectionTimeout;
    }

    public void setConnectionTimeout(long connectionTimeout) {
        this.connectionTimeout = connectionTimeout;
    }

    public long getReceiveTimeout() {
        return receiveTimeout;
    }

    public void setReceiveTimeout(long receiveTimeout) {
        this.receiveTimeout = receiveTimeout;
    }

    public String getSnapCenterBaseURL() {
        return snapCenterBaseURL;
    }

    public void setSnapCenterBaseURL(String snapCenterBaseURL) {
        this.snapCenterBaseURL = snapCenterBaseURL;
    }

}